(******************************************************************************)
(*  ocsigen-bundler: create self-contained Ocsigen application package        *)
(*                                                                            *)
(*  Copyright (C) 2011, OCamlCore SARL                                        *)
(*                                                                            *)
(*  This library/application is free software; you can redistribute it        *)
(*  and/or modify it under the terms of the GNU Lesser General Public         *)
(*  License as published by the Free Software Foundation; either version 2.1  *)
(*  of the License, or (at your option) any later version, with the OCaml     *)
(*  static compilation exception.                                             *)
(*                                                                            *)
(*  This library/application is distributed in the hope that it will be       *)
(*  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file         *)
(*  COPYING for more details.                                                 *)
(*                                                                            *)
(*  You should have received a copy of the GNU Lesser General Public License  *)
(*  along with this library; if not, write to the Free Software Foundation,   *)
(*  Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA             *)
(******************************************************************************)

open Xml
open ExtLib

type t = 
    {
      modules          : string list;
      findlib_packages : string list;
      copies           : (string * string) list;
      logdir_set       : bool;
      commandpipe_set  : bool;
      port_set         : bool;
    }

let () =
  (* Determine if ld.conf is overriden env variables. *)
  begin
    try 
      if Sys.getenv "OCAMLFIND_LDCONF" = "ignore" then
        (* Acting on the UNIX env, that is bad for a library. *)
        (* TODO: fix this. *)
        Unix.putenv "OCAMLFIND_LDCONF" ""
    with Not_found ->
      ()
  end;
  Findlib.init ()

exception Warning of string 

module SetRootPkg = Set.Make(String)
module SetString = Set.Make(String)

let info ?(verbose=false) fmt = 
  if verbose then
    begin
      Printf.printf "I: ";
      Printf.kfprintf
        (fun chn -> Printf.fprintf chn "\n%!")
        stdout
        fmt
    end
  else
    begin
      Printf.ifprintf stdout fmt
    end

let warning ?(verbose=false) fmt = 
  Printf.eprintf "W: ";
  Printf.kfprintf 
    (fun chn -> Printf.fprintf chn "\n%!")
    stderr
    fmt

let deploy ?verbose ?(dry_run=false) ?host ~tgt src = 
  let info fmt = info ?verbose fmt in
  let warning fmt = warning ?verbose fmt in

  let start = 
    Printf.sprintf "cd %s && ./daemon.sh start" tgt
  in
  let stop = 
    Printf.sprintf "cd %s && ./daemon.sh stop" tgt
  in
  let rsync = 
    Printf.sprintf "rsync -avP %s/* %s" src tgt
  in

  let start, stop, rsync, clean_exit = 
    match host with 
      | Some hst ->
          let tmpfn = 
            Filename.temp_file "ocsigen-bundler" ".ssh-master"
          in
          let mstr_process = 
            (* Start SSH in master mode *)
            let cmd = 
              Printf.sprintf 
                "ssh -M -S %s %s 'read'" tmpfn hst 
            in
              FileUtil.rm [tmpfn];
              info "Running command '%s'" cmd;
              if not dry_run then
                Unix.open_process_out cmd
              else
                Unix.open_process_out "true"
          in
          let clean_exit () = 
            output_string mstr_process "\n";
            match Unix.close_process_out mstr_process with 
              | Unix.WEXITED 0 -> 
                  ()
              | _ -> 
                  warning 
                    "Error while closing connection to %s"
                    hst
          in
          let start = 
            Printf.sprintf "ssh -S %s %s '%s'" tmpfn hst start
          in
          let stop = 
            Printf.sprintf "ssh -S %s %s '%s'" tmpfn hst stop
          in
          let rsync = 
            Printf.sprintf "rsync -e 'ssh -t -S %s' -av %s/* %s:%s"
              tmpfn src hst tgt
          in
            start, stop, rsync, clean_exit

      | None ->
          start, stop, rsync, ignore
  in

  let exec cmd =
    info "Running command '%s'" cmd;
    if not dry_run then
      Sys.command cmd
    else
      0
  in

  let exec_raise cmd =
    match exec cmd with
      | 0 -> ()
      | n ->
          failwith 
            (Printf.sprintf 
               "Command '%s' exited with code %d"
               cmd n)
  in

    try 
      begin
        match exec stop with 
          | 0 ->
              ()
          | n ->
              warning "Command '%s' exited with code %d" stop n
      end;
      exec_raise rsync;
      exec_raise start;
      clean_exit ()
    with e ->
      clean_exit ();
      raise e

let make ?verbose ?(dry_run=false) ?port ?logdir ?(mkdir=[]) ?(copy_dir=[]) ?(copy_file=[]) ?(natives=[]) fn tgt = 
  let info fmt = info ?verbose fmt in
  let warning fmt = warning ?verbose fmt in

  let bindir = 
    "bin"
  in

  let std_stublib_dirs =
    (* Location of stub libraries *)
    []
  in

  let logdir, mkdir_logdir = 
    match logdir with 
      | Some s ->
          s, false
      | None ->
          "log", true
  in

  let etcdir = 
    "etc"
  in

  let moddir = 
    "mod"
  in

  let rundir = 
    "run"
  in

  let libdir = 
    "lib"
  in
  
  let ocamldir =
    FilePath.concat libdir "ocaml"
  in
    
  let _metasdir =
    FilePath.concat ocamldir "METAS"
  in

  let stubdir =
    FilePath.concat ocamldir "stublibs"
  in

  let commandpipe =
    FilePath.concat rundir "ocsigen_command"
  in

  let rec partition_one nm lst =
    match lst with 
      | (nm', res) :: tl when nm' = nm ->
          res, tl

      | e :: tl ->
          let res , tl =
            partition_one nm tl
          in
            res, e :: tl

      | [] ->
          raise Not_found
  in

  let rec parse t =
    function
      | Element ("extension", attrs, children) as xml ->
          begin
            try
              let fnd_pkg = 
                List.assoc "findlib-package" attrs
              in
                xml,
                {t with 
                  findlib_packages = fnd_pkg :: t.findlib_packages}

            with Not_found ->
              failwith "Unhandle extension element"
          end

      | Element ("eliom", attrs, children) ->
          begin
            try 
              let modl, attrs = 
                partition_one "module" attrs
              in
              let new_modl = 
                FilePath.make_filename 
                  [moddir; FilePath.basename modl]
              in
                Element ("eliom",
                         (("module", new_modl) :: attrs),
                         children),
                {t with 
                  modules = modl :: t.modules;
                  copies  = (modl, new_modl) :: t.copies }

            with Not_found ->
              failwith "Unhandle eliom element"
          end

      | Element ("logdir", attrs, children) ->
          Element ("logdir", attrs, [PCData logdir]),
          {t with logdir_set = true}

      | Element ("commandpipe", attrs, children) ->
          Element ("commandpipe", attrs, [PCData commandpipe]),
          {t with commandpipe_set = true}

      | Element ("port", attrs, children) as e ->
          begin
            match port with 
              | Some n ->
                  Element("port", attrs, [PCData (string_of_int n)]),
                  {t with port_set = true}
              | None ->
                  e, t
          end

      | Element (nm, attrs, children) as xml ->
          let children, t = 
            fold 
              (fun (acc, t) xml ->
                 let e, t = 
                   parse t xml
                 in
                   e :: acc, t)
              ([], t) 
              xml
          in
          let children =  
            List.rev children
          in
          let children = 
            if nm = "server" then
              begin
                let children = 
                  List.fold_left
                    (fun acc (tst, nm, vl) ->
                       if not tst then
                         Element (nm, [], [PCData vl]) :: acc
                       else
                         acc)
                    children
                    [
                      t.logdir_set, "logdir", logdir;
                      t.commandpipe_set, "commandpipe", commandpipe;
                    ]
                in
                let children =
                  match t.port_set, port with
                    | false, Some n ->
                        (Element ("port", [], [PCData (string_of_int n)])) 
                        :: 
                        children
                    | _, _ ->
                        children
                in
                  children
              end
            else
              children
          in
            Element (nm, attrs, children),
            t

      | PCData str as pcdata ->
          pcdata, t
  in

  let xml, t = 
    parse 
      {
        modules          = [];
        findlib_packages = [];
        copies           = [];
        port_set         = false;
        commandpipe_set  = false;
        logdir_set       = false;
      } 
      (Xml.parse_file fn)
  in

  let predicates = 
    ["byte"; "mt"]
  in

  let all_findlib_packages =
    (* Findlib packages + there ancestor *)
    Findlib.package_deep_ancestors 
      predicates 
      t.findlib_packages
  in

  let findlib_copies = 
    let pkg_prop pkg prop = 
      try 
        Findlib.package_property 
          predicates 
          pkg 
          prop
      with Not_found ->
        raise 
          (Warning
             (Printf.sprintf 
                "Findlib property '%s' not found for package '%s'" 
                prop pkg))
    in

    let to_root_pkg pkg = 
      try 
        fst (String.split pkg ".")
      with Invalid_string ->
        pkg
    in

    let to_meta root_pkg = 
      try 
        let lst = 
          List.fold_left 
            (fun acc dn ->
               (true, FilePath.concat dn ("META."^root_pkg))
               ::
               (false, FilePath.make_filename [dn; root_pkg; "META"])
               ::
               acc)
            []
            (Findlib.search_path ())
        in
          List.find 
            (fun (_, fn) -> Sys.file_exists fn) 
            (List.rev lst)
      with Not_found -> 
        failwith 
          (Printf.sprintf 
             "META file not found for package '%s'"
             root_pkg)
    in

    let to_archive pkg = 
      let archives = 
        String.nsplit 
          (pkg_prop pkg "archive")
          " "
      in
      let archives = 
        List.filter (( <> ) "") archives
      in
        archives
    in


    let copies = 
      []
    in

    let copies, _ = 
      (* Copy META files *)
      List.fold_left
        (fun (lst, st) pkg ->
           let root_pkg = 
             to_root_pkg pkg
           in
             if not (SetRootPkg.mem root_pkg st) then
               begin
                 let is_outdir, meta = 
                   to_meta root_pkg
                 in
                 let bn = 
                   FilePath.basename meta
                 in
                 let tgt = 
                   if is_outdir then
                     FilePath.concat "METAS" bn 
                   else
                     FilePath.concat root_pkg bn
                 in
                   (meta, tgt) :: lst,
                   SetRootPkg.add root_pkg st
               end
             else
               lst, st)
        (copies, SetRootPkg.empty)
        all_findlib_packages
    in

    let copies = 
      (* Copy archive files *)
      List.fold_left
        (fun acc pkg ->
           try 
             let archives = 
               to_archive pkg
             in
               
             let location =
               Findlib.package_directory pkg
             in

             let to_tgt =
               if location = Findlib.ocaml_stdlib () then
                 (* This is a standard library *)
                 fun s -> s
               else
                 FilePath.concat (to_root_pkg pkg)
             in
               List.fold_left
                 (fun acc fn ->
                    (FilePath.concat location fn, to_tgt fn)
                    ::
                    acc)
                 acc
                 archives
        
           with Warning msg ->
             warning "%s" msg;
             acc)
        copies
        all_findlib_packages 
    in

    let copies = 
      (FilePath.make_filename [Findlib.ocaml_stdlib (); "threads"; "threads.cma"],
       FilePath.concat "threads" "threads.cma")
      ::
      (FilePath.make_filename [Findlib.ocaml_stdlib (); "vmthreads"; "threads.cma"],
       FilePath.concat "vmthreads" "threads.cma")
      :: copies
    in

      copies
  in

  let install ?tgt_rel (src, tgt_tl) = 
    let tgt_fn = 
      match tgt_rel with 
        | Some tgt_rel ->
            FilePath.make_filename [tgt; tgt_rel; tgt_tl]
        | None ->
            FilePath.concat tgt tgt_tl
    in
      info "install '%s' -> '%s'" src tgt_fn;
      if not dry_run then
        begin
          FileUtil.mkdir ~parent:true (FilePath.dirname tgt_fn);
          FileUtil.cp [src] tgt_fn
        end
  in

  let info_mkdir dn = 
    let dn =
      FilePath.concat tgt dn
    in
      info "mkdir '%s'" dn;
      if not dry_run then
        FileUtil.mkdir ~parent:true dn
  in

  let install_native cmd = 
    install ~tgt_rel:bindir (FileUtil.which cmd, cmd);
    if not dry_run then 
      Unix.chmod (FilePath.make_filename [tgt; bindir; cmd]) 0o755
  in

  let dll_copies = 
     (* We only care about dll for stub library (OCaml/C bindins found in
      * stublibs/). We will use Ocsigen_ldd to determine dependencies of this 
      * stub libraries (solved by ldd, so the right one).
      *)
    let extract_lopt racc str = 
      try
        Array.iter
          (function 
             | [| _; hd |] ->
                 racc := SetString.add hd !racc
             | _ ->
                 ())
          (Pcre.extract_all ~pat:"-l([^ ]+)" str)
      with Not_found ->
        ()
    in

    let extract_so racc str = 
      try
        Array.iter
          (function 
             | [| _; hd |] ->
                 racc := SetString.add hd !racc
             | _ ->
                 ())
          (Pcre.extract_all ~pat:"dll([^ ]+)\\.so" str)
      with Not_found ->
        ()
    in

    let to_dll acc_dllstub cma =
      let cmd_line = 
        Printf.sprintf "ocamlobjinfo '%s'" cma
      in
      let chn =
        Unix.open_process_in cmd_line
      in
      let racc_dllstub = ref acc_dllstub in
      let () = 
        try 
          while true do 
            let line = 
              input_line chn
            in
              try 
                match ExtString.String.split line ":" with
                  | "Extra dynamically-loaded libraries", value ->
                      extract_lopt racc_dllstub value;
                      extract_so racc_dllstub value
                  | fld, value ->
                      ()
              with ExtLib.Invalid_string ->
                ()
          done
        with End_of_file ->
          ()
      in
      let () = 
        match Unix.close_process_in chn with 
          | Unix.WEXITED 0 -> 
              ()
          | _ -> 
              failwith 
                (Printf.sprintf 
                   "Command '%s' exited with non zero status code." 
                   cmd_line)
      in
        !racc_dllstub
    in

    let stublib_dirs =
      let common_dirs = 
        match Findlib.ocaml_ldconf () with 
          | "ignore" ->
              []
          | fn ->
              let chn = open_in fn in
              let lst = ref [] in
              let () =
                try 
                  while true do 
                    lst := (input_line chn) :: !lst
                  done;
                with End_of_file ->              
                  close_in chn
              in
                List.rev !lst
      in
      let pkg_dirs = 
        List.map Findlib.package_directory all_findlib_packages 
      in
        common_dirs @ pkg_dirs @ std_stublib_dirs
    in

    let locate_dll ~where nm = 
      let dll_basename = Printf.sprintf "dll%s.so" nm in
        try 
          List.find 
            Sys.file_exists 
            (List.map 
               (fun dn -> Filename.concat dn dll_basename)
               where)
        with Not_found ->
          failwith 
            (Printf.sprintf
               "Unable to find dll stub library '%s'"
               dll_basename)
    in

    let cma = 
      List.filter 
        (fun fn -> 
           FilePath.check_extension fn "cma"
           ||
           FilePath.check_extension fn "cmo")
        (List.map fst (List.rev_append t.copies findlib_copies))
    in

    let dllstub = 
      List.fold_left to_dll SetString.empty cma
    in

    let dllstub_fn = 
      SetString.fold 
        (fun nm acc -> locate_dll ~where:stublib_dirs nm :: acc)
        (SetString.add "threads" dllstub)
        []
    in

      List.map (fun fn -> fn, FilePath.basename fn) dllstub_fn
  in

  let copy_dir = 
    (* TODO: remove when bug 
     * https://forge.ocamlcore.org/tracker/index.php?func=detail&aid=814&group_id=128&atid=589
     * will be fixed.
     *)
    List.fold_left
      (fun acc (dir, tgt) ->
         let mk_abs = 
           FilePath.make_absolute
             (FileUtil.pwd ())
         in
         let dir_abs =
           mk_abs dir
         in
           FileUtil.find 
             FileUtil.Is_file  
             dir_abs
             (fun acc fn ->
                let tgt_fn = 
                  FilePath.reparent dir_abs (mk_abs tgt) fn
                in
                let tgt_fn =
                  FilePath.make_relative (FileUtil.pwd ()) tgt_fn
                in
                  (fn, tgt_fn) :: acc)
             acc)
      []
      copy_dir
  in

  let natives = 
    "ocsigen" :: "ocamlrun" :: natives
  in

  let extra_dll_copies = 
    let bins = 
      List.map FileUtil.which natives
    in
    let libs = 
      List.map fst dll_copies 
    in
    let libs_included = 
      List.fold_left 
        (fun st fn -> SetString.add fn st) 
        SetString.empty 
        libs
    in
      (* Get .so dependencies + dedup with dll_copies *)
      List.filter
        (fun (fn, _) -> not (SetString.mem fn libs_included))
        (Ocsigen_ldd.ldd (List.rev_append bins libs))
  in

    if mkdir_logdir then 
      info_mkdir logdir;

    info_mkdir rundir;
    info_mkdir etcdir;
    List.iter info_mkdir mkdir;

    List.iter install t.copies;
    List.iter install_native natives;
    List.iter (install ~tgt_rel:ocamldir) findlib_copies;
    List.iter (install ~tgt_rel:stubdir) dll_copies;
    List.iter (install ~tgt_rel:libdir) extra_dll_copies;

    begin
      let fn = 
        FilePath.make_filename [tgt; etcdir; "ocsigen.conf"]
      in
      let ctnt = 
        Xml.to_string_fmt xml
      in
        info "Creating file '%s'" fn;
        List.iter 
          (fun s -> info "%s" s)
          (String.nsplit ctnt "\n");
        if not dry_run then
          let chn = 
            open_out fn
          in
            output_string chn ctnt;
            close_out chn
    end;

    begin
      let fn =
        FilePath.concat tgt "daemon.sh"
      in
        info "Creating file '%s'" fn;
        if not dry_run then
          let chn =
            open_out fn
          in
            output_string 
              chn
              Ocsigen_bundler_data.daemon_sh;
            close_out chn;
            Unix.chmod fn 0o755
    end;

    List.iter install copy_dir;
    List.iter install copy_file;

    ()

